//+------------------------------------------------------------------+
//|                                               mnist_original.mq5 |
//|                                  Copyright 2023, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2023, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#property version   "1.00"

#include "NeuralNetworkClassification.mqh"
#include "..\mnist_data_original\MNIST_data.mqh"

input uint InpImagesOut=0;    // number of pictures to print into log

struct data_one
  {
   vector            image;
   int               label;
  };

//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
  {
   int res=mnist();
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int ReadData(data_one& data[],const uchar& labels_array[],const uchar& images_array[])
  {
   LabelHeader label_header;
   ImageHeader image_header;
   uint        images;
   uint        items;
   uint        rows;
   uint        cols;
   int         copied=ArrayCopy(label_header.char_array,labels_array,0,0,sizeof(LabelHeaderStruct));
   if(copied!=sizeof(LabelHeaderStruct))
      return(-1);
   items=MathSwap(label_header.header.items);

   ArrayResize(data,items);
   for(unsigned int i=0; i < items; i++)
      data[i].label=labels_array[i+sizeof(LabelHeaderStruct)];

   copied=ArrayCopy(image_header.char_array,images_array,0,0,sizeof(ImageHeaderStruct));
   if(copied!=sizeof(ImageHeaderStruct))
      return(-2);
   images=MathSwap(image_header.header.images);
   rows=MathSwap(image_header.header.rows);
   cols=MathSwap(image_header.header.cols);
   if(items!=images || rows!=28 || cols!=28)
      return(-3);

   uint   cnt=0;
   string out_line;
   for(uint i=0; i<images; i++)
     {
      if(cnt<InpImagesOut)
         printf("\nimage %d",data[i].label);
      data[i].image.Init(28*28);
      for(int j=0; j<28*28; j++)
        {
         int color_value=int(images_array[i*28*28+sizeof(ImageHeaderStruct)+j]);
         data[i].image[j] = double(color_value)/255.0;
         if(cnt<InpImagesOut)
           {
            out_line+=IntegerToString(color_value,3)+" ";
            if(j%28==0)
              {
               Print(out_line);
               out_line="";
              }
           }
        }
      if(cnt<InpImagesOut)
         Print("");
      cnt++;
     }

   return(int(images));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int mnist()
  {
//--- read train data
   data_one data[];
   int      images=ReadData(data,ExtTrainLabels,ExtTrainImages);
   if(images<0)
      return(images);
   printf("Read %d images. Learning begin",images);

//--- create   neural network
   srand(uint(TimeLocal()));

   const int                layers=3;
   int                      sizes[3]={28*28, 128, 10};
   ENUM_ACTIVATION_FUNCTION act[3]={AF_NONE, AF_RELU, AF_SOFTMAX};
   const double             learning_rate=0.05;
   CNeuralNetwork nn(layers,sizes,act);

//--- training
   const int epochs_cnt=15;
   vector    vector_pred;              // predicted class probabilities
   vector    vector_true(10);          // one hot encoded right answer
   uint      start=GetTickCount();

   for(int e=0; e<epochs_cnt; e++)
     {
      int  right_answers=0;
      uint epoch_start=GetTickCount();
      printf("Epoch # %d",e+1);

      for(int i=0; i<images; i++)
        {
         if(IsStopped())
           {
            Print("stopped");
            return(-10);
           }
         nn.ForwardFeed(data[i].image,vector_pred);
         int prediction_index=(int)vector_pred.ArgMax();

         if(prediction_index==data[i].label)
            right_answers++;
         else
           {
            vector_true.Fill(0.0);
            vector_true[data[i].label]=1.0; // one hit
            nn.BackPropagation(vector_true,learning_rate);
           }
        }

      uint   epoch_stop=GetTickCount();
      double epoch_duration=(double)(epoch_stop-epoch_start)/1000.0;
      printf("Time: %.3f seconds.\tRight answers: %.3f%%",epoch_duration,(right_answers/double(images)) * 100.0);
     }

   printf("Total time: %d seconds",(GetTickCount()-start)/1000);

//--- read test data
   data_one test_data[];
   images=ReadData(test_data,ExtTestLabels,ExtTestImages);
   if(images<0)
      return(images-3);
   printf("Read %d images. Test begin",images);

//--- testing
   int    right_answers=0;
   vector test_result;
   uint   test_start=GetTickCount();

   for(int test=0; test<images; test++)
     {
      nn.ForwardFeed(test_data[test].image,test_result);
      int prediction_index=(int)test_result.ArgMax();
      if(prediction_index==test_data[test].label)
         right_answers++;
     }

   uint   test_stop=GetTickCount();
   double test_duration=(double)(test_stop-test_start)/1000.0;
   printf("Time: %.3f seconds.\tRight answers: %.3f%%",test_duration,(double(right_answers)/double(images))*100.0);

//--- ok
   return(0);
  }
//+------------------------------------------------------------------+
